//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Detector/Detector2D.cpp
//! @brief     Implements class Detector2D.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Detector/Detector2D.h"
#include "Base/Axis/Frame.h"
#include "Base/Axis/MakeScale.h"
#include "Base/Axis/Scale.h"
#include "Base/Pixel/Pixel.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/Beam.h"
#include "Device/Resolution/IDetectorResolution.h"

namespace {

//! Returns true scattering angles for given nominal scale.
//! The nominal scale is an equidistant partition of the interval between true minimum and
//! maximum scattering angle. It is used for the graphical representation of detector pixels.
//! The true scattering angle is used for the computation of scattering towards a detector pixel.
Scale* scatteringAngles(const Scale& in)
{
    const size_t n = in.size();
    ASSERT(n > 0);
    const double p0 = tan(in.min());
    const double dp = (tan(in.max()) - tan(in.min())) / n;
    std::vector<double> limits;
    limits.reserve(2 * n);
    limits.push_back(in.min());
    for (size_t i = 1; i < n; ++i) {
        limits.push_back(atan(p0 + i * dp));
        limits.push_back(atan(p0 + i * dp));
    }
    limits.push_back(in.max());
    return newGenericScale(in.axisLabel(), limits);
}

} // namespace


Detector2D::Detector2D(size_t n_phi, double phi_min, double phi_max, size_t n_alpha,
                       double alpha_min, double alpha_max)
    : IDetector(new Frame(newEquiDivision("phi_f (rad)", n_phi, phi_min, phi_max),
                          newEquiDivision("alpha_f (rad)", n_alpha, alpha_min, alpha_max)))
{
    precompute();
}

Detector2D::Detector2D(const Detector2D& other)
    : IDetector(other)
{
    precompute();
}

Detector2D* Detector2D::clone() const
{
    return new Detector2D(*this);
}

void Detector2D::precompute()
{
    m_scattering_phi.reset(::scatteringAngles(frame().xAxis()));
    m_scattering_alpha.reset(::scatteringAngles(frame().yAxis()));
}


Pixel* Detector2D::createPixel(size_t index) const
{
    const size_t phi_index = axisBinIndex(index, 0);
    const size_t alpha_index = axisBinIndex(index, 1);

    const Bin1D alpha_bin = m_scattering_alpha->bin(alpha_index);
    const Bin1D phi_bin = m_scattering_phi->bin(phi_index);
    return new Pixel(alpha_bin, phi_bin);
}

size_t Detector2D::indexOfSpecular(const Beam& beam) const
{
    double alpha = beam.alpha_i();
    double phi = beam.phi_i();
    if (m_scattering_phi->rangeComprises(phi) && m_scattering_alpha->rangeComprises(alpha))
        return getGlobalIndex(m_scattering_phi->closestIndex(phi),
                              m_scattering_alpha->closestIndex(alpha));
    return totalSize();
}
